from ..types.plotstyle import PlotStyle

#
# Constants
#

style_area = PlotStyle()
style_areabr = PlotStyle()
style_circles = PlotStyle()
style_columns = PlotStyle()
style_cross = PlotStyle()
style_histogram = PlotStyle()
style_line = PlotStyle()
style_linebr = PlotStyle()
style_stepline = PlotStyle()
style_stepline_diamond = PlotStyle()
